# %% [markdown]
# Simple ISBM Pub/Sub Example

# %%
print('Simple ISBM Pub/Sub Example')

# %% [markdown]
# Used Libraries
# * Requests: [https://requests.readthedocs.io/](https://requests.readthedocs.io/)
# * openapi-core: [https://github.com/p1c2u/openapi-core](https://github.com/p1c2u/openapi-core)
# 
# Ensure the libraries are installed with:
# ```
# pip install requests openapi-core
# ```

# %%
import requests
from urllib.parse import quote as urlquote, urlparse
from openapi_core import create_spec
from openapi_core.validation.request.validators import RequestValidator, RequestValidationResult
from openapi_core.validation.response.validators import ResponseValidator, ResponseValidationResult
from openapi_core.contrib.requests import RequestsOpenAPIRequest, RequestsOpenAPIResponse
import yaml

# %%
# load the spec file and read the yaml/json
# (download the file from OpenO&M website 'https://www.openoandm.org/isbm/2.0/openapi/isbm_complete.json')
with open('isbm_spec.json') as f:
   spec_dict = yaml.safe_load(f.read())

# or load from the web endpoint (this example uses the ISBM instance for the OIIE Australian Working Group)
# (Note: at time of recording the endpoint is distributing an out-of-date version of spec)
# import urllib.request
# with urllib.request.urlopen('https://isbm.au-wg.oiiecosystem.net/rest/api') as content:
    # spec_dict = yaml.safe_load(content.read())
with requests.get('https://isbm.au-wg.oiiecosystem.net/rest/api') as response:
    response.raise_for_status()
    spec_dict = response.json()

host_url = 'https://isbm.au-wg.oiiecosystem.net/rest'
spec_dict['servers'] = [{'url': host_url}]
spec_dict['info']

# %%
# "Fix" path patterns for Python syntax conformance
# Internally, openapi-core turns the URL parameters into object attributes, 
# which causes path matching to fail if the parameter does not conform to Python syntax.
import re

path_dict = spec_dict['paths']
for path in [p for p in path_dict.keys() if re.search('[{][^}-]+[-][^}]+[}]', p) ]:
    for param in [p for p in path_dict[path]['parameters'] if '-' in p['name']]:
        param['name'] = param['name'].replace('-', '_')
    safe_path = re.sub('[{][^}-]+[-][^}]+[}]', lambda m: m.group().replace('-', '_')  , path)
    path_dict[safe_path] = path_dict[path]
    del path_dict[path]


# %%
# Fix optional security schemes
spec_dict['security'].append({})

# %%
# from jsonpath_ng.ext import parse

# The openapi schema validator is really picky with the 'floats' used in the 'minItems'
# match_exp = parse('$..*[?(minItems)]')
# for match in match_exp.find(spec_dict):
#     match.value['minItems'] = int(match.value['minItems'])


# %%
# Custom format checkers for validation
import re

class UriFormatter:
    def validate(self, value) -> bool:
        try:
            urlparse(value) # Does it always raise?
            return True
        except:
            return False

class DurationFormatter:
    _iso8106_format = re.compile('[-]?P([0-9]+Y)?([0-9]+M)?([0-9]+D)?(T([0-9]+H)?([0-9]+M)?([0-9]+([.][0-9]+)?S)?)?')
    
    def validate(self, value) -> bool:
        return bool(self._iso8106_format.match(value))
    
    def unmarshal(self, value) -> str:
        return value


# %%
# parse the spec into python - this will raise if the spec is invalid
spec = create_spec(spec_dict)

custom_formatters = { 'uri': UriFormatter(), 'duration': DurationFormatter() }
request_validator = RequestValidator(spec, base_url=host_url, custom_formatters=custom_formatters)
response_validator = ResponseValidator(spec, base_url=host_url, custom_formatters=custom_formatters)

# %%
# Common values for pub/sub
channel_uri = '/example/hello_world/publication'
topic = 'echo-example'
text_message_content = 'Hello World!'
json_message_content = { 'say': 'Hello World!' }
xml_message_content = '<say>Hello World!</say>'

# %%
# Helpers functions

from jsonpath_ng.ext import parse
from typing import Dict

def get_operation_path(operation_name: str, path_elements: Dict[str,str]={}) -> str:
    """Retrieves the relative URL path for the given ISBM operation from the OpenAPI spec.
    Provided path elements are used to replace URL parameters if provided and they match
    the URL parameter names. Path elements will be percent encoded, including channel URIs.
    That is, channel URIs will have the '/' encoded to '%2F' to form a single path element.

    Args:
        operation_name (str): the name of the operation
        path_elements (Dict[str,str], optional): URL parameter variables (keys) and their values. 
            Defaults to {}.

    Returns:
        str: relative URL path component with URL parameters replaced if provided
    """
    operation_path_exp = parse('$..*[?(operationId="%s")]' % operation_name)
    paths = operation_path_exp.find(spec_dict)
    if len(paths) == 0:
        return ''
    else:
        path = paths[0].full_path.left.right.fields[0]
        for name, value in path_elements.items():
            path = path.replace('{' + name + '}', urlquote(value, safe=''))
        return path

def get_operation_method(operation_name: str) -> str:
    """Retreives the HTTP method for the given ISBM operation from the OpenAPI spec
    for use in configuring requests to the server. That is the method is return uppercase.

    Args:
        operation_name (str): the name of the operation

    Returns:
        str: HTTP method name (uppercase)
    """
    operation_path_exp = parse('$..*[?(operationId="%s")]' % operation_name)
    paths = operation_path_exp.find(spec_dict)
    if len(paths) == 0:
        return 'NONE'
    else:
        operation_dict = paths[0].full_path.left.find(spec_dict)[0].value
        method = [k for k, v in operation_dict.items() if isinstance(v, dict) and v['operationId'] == operation_name][0]
        return method.upper()

def validate_request(request: requests.PreparedRequest, validator: RequestValidator=request_validator) -> RequestValidationResult:
    """Validates the given request with respect to the OpenAPI spec validator, which defaults to 
    the preconfigured validator; raises validation errors as exceptions if validation fails.
    We raise errors to halt the code execution in the cell.

    Args:
        request (requests.PreparedRequest): the (prepared) request that requires validation
        validator (RequestValidator, optional): Defaults to request_validator.

    Returns:
        RequestValidationResult: the validation result
    """
    openapi_request = RequestsOpenAPIRequest(request)
    result = validator.validate(openapi_request)
    result.raise_for_errors()
    return result

def validate_response(response: requests.Response, validator: ResponseValidator=response_validator) -> ResponseValidationResult:
    """Validates the given response with respect to the OpenAPI spec validator, which defaults to
    the preconfigured validator; reaises validation errors as exceptions if the validation fails.
    We raise errors to halt the code execution in the cell.

    Args:
        response (requests.Response): the response requiring validation
        validator (ResponseValidator, optional): Defaults to response_validator.

    Returns:
        ResponseValidationResult: the validation result
    """
    openapi_response = RequestsOpenAPIResponse(response)
    result = validator.validate(RequestsOpenAPIRequest(response.request), openapi_response)
    result.raise_for_errors()
    return result

# %%
# Test it is working by retrieving all (open) channels
response = requests.get(host_url + get_operation_path('getChannels'))
response.raise_for_status() # will raise an exception if not success
response.json()

# %%
# Validate the response w.r.t. the OpenAPI spec
openapi_response = RequestsOpenAPIResponse(response)
result = response_validator.validate(RequestsOpenAPIRequest(response.request), openapi_response)
result.raise_for_errors()

# %%
# prepare a session to execute requests
requests_session = requests.Session()
requests_session.headers['Accept'] = 'application/json'
# requests_session.auth = ('username', 'password')

# %%
# Open Publish Session
req = requests.Request(get_operation_method('openPublicationSession'), host_url + get_operation_path('openPublicationSession', {'channel_uri': channel_uri}))
prepared_req = requests_session.prepare_request(req)
validate_request(prepared_req)

response = requests_session.send(prepared_req)
response.raise_for_status()
validate_response(response)

publish_session_id = response.json()['sessionId']
print('publish_session_id:', publish_session_id)

# %%
# Open Subscribe Session
req = requests.Request(get_operation_method('openSubscriptionSession'), host_url + get_operation_path('openSubscriptionSession', {'channel_uri': channel_uri}))
req.json = { 'topics': [ topic ] }
prepared_req = requests_session.prepare_request(req)
validate_request(prepared_req)

response = requests_session.send(prepared_req)
response.raise_for_status()
validate_response(response)

subscribe_session_id = response.json()['sessionId']
print('subscribe_session_id:', subscribe_session_id)

# %%
# Publish Message (String)
req = requests.Request(get_operation_method('postPublication'), host_url + get_operation_path('postPublication', {'session_id': publish_session_id}))
req.json = { 
    'messageContent': { 
        'mediaType': 'text/plain', 
        'content': text_message_content 
    }, 
    'topics': [ topic ],
    'expiry': 'PT24H' # optional expiry value, ISO 8601 Duration, e.g., 24 hours
}
prepared_req = requests_session.prepare_request(req)
validate_request(prepared_req)

response = requests_session.send(prepared_req)
response.raise_for_status()
validate_response(response)

message_id = response.json()['messageId']
print('posted message, id:', message_id)

# %%
# Publish JSON Message
req = requests.Request(get_operation_method('postPublication'), host_url + get_operation_path('postPublication', {'session_id': publish_session_id}))
req.json = { 
    'messageContent': { 
        'content': json_message_content
    }, 
    'topics': [ topic ],
    'expiry': 'PT24H' # optional expiry value, ISO 8601 Duration, e.g., 24 hours
}
prepared_req = requests_session.prepare_request(req)
validate_request(prepared_req)

response = requests_session.send(prepared_req)
response.raise_for_status()
validate_response(response)

message_id = response.json()['messageId']
print('posted message, id:', message_id)

# %%
# Publish XML Message
req = requests.Request(get_operation_method('postPublication'), host_url + get_operation_path('postPublication', {'session_id': publish_session_id}))
req.json = { 
    'messageContent': { 
        'mediaType': 'application/xml',
        'content': xml_message_content
    }, 
    'topics': [ topic ],
    'expiry': 'PT24H' # optional expiry value, ISO 8601 Duration, e.g., 24 hours
}
prepared_req = requests_session.prepare_request(req)
validate_request(prepared_req)

response = requests_session.send(prepared_req)
response.raise_for_status()
validate_response(response)

print('posted message, id:', response.json()['messageId'])

# %%
# Read Message
req = requests.Request(get_operation_method('readPublication'), host_url + get_operation_path('readPublication', {'session_id': subscribe_session_id}))
prepared_req = requests_session.prepare_request(req)
validate_request(prepared_req)

response = requests_session.send(prepared_req)
response.raise_for_status() if response.status_code != 200 and response.status_code != 404 else None
validate_response(response)

message = response.json()
print('read message', message)

# %%
# Remove Message
req = requests.Request(get_operation_method('removePublication'), host_url + get_operation_path('removePublication', {'session_id': subscribe_session_id}))
prepared_req = requests_session.prepare_request(req)
validate_request(prepared_req)

# 204 No Content if successful (even if no more messages to remove), so no need to validate the response
response = requests_session.send(prepared_req)
print('Message removed' if response.status_code == 204 else 'There was an error')


# %%
# Expire the JSON Message
req = requests.Request(get_operation_method('expirePublication'), 
                        host_url + get_operation_path('expirePublication', {'session_id': publish_session_id, 'message_id': message_id}))
prepared_req = requests_session.prepare_request(req)
validate_request(prepared_req)

# 204 No Content if successful (even if message no longer around to be expired), so no need to validate the response
response = requests_session.send(prepared_req)
print('Message expired' if response.status_code == 204 else 'There was an error')

# %%
# Close publication session
req = requests.Request(get_operation_method('closeSession'), host_url + get_operation_path('closeSession', {'session_id': publish_session_id}))
prepared_req = requests_session.prepare_request(req)
response = requests_session.send(prepared_req)
response.raise_for_status()
print(response)

# %%
# Close subscribe session
req = requests.Request(get_operation_method('closeSession'), host_url + get_operation_path('closeSession', {'session_id': subscribe_session_id}))
prepared_req = requests_session.prepare_request(req)
response = requests_session.send(prepared_req)
response.raise_for_status()
print(response)

# %%
# Really should close the shared HTTP session
requests_session.close()


